@@ -21,28 +21,27 @@
  *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.          *
  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
+#include <SDL.h>
+#include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include <SDL.h>
-
 #define M64P_PLUGIN_PROTOTYPES 1
-#include "m64p_types.h"
-#include "m64p_plugin.h"
+#include "config.h"
 #include "m64p_common.h"
 #include "m64p_config.h"
-
-#include "plugin.h"
-#include "config.h"
-#include "version.h"
+#include "m64p_plugin.h"
+#include "m64p_types.h"
 #include "osal_dynamiclib.h"
+#include "plugin.h"
+#include "version.h"
 
 #ifdef __linux__
-#include <unistd.h>
 #include <dirent.h>
 #include <fcntl.h>
 #include <sys/types.h>
+#include <unistd.h>
 #include <linux/input.h>
 #endif /* __linux__ */
 
@@ -111,7 +110,7 @@ static int romopen = 0;         // is a rom opened
 
 static unsigned char myKeyState[SDL_NUM_SCANCODES];
 
-#ifdef __linux__
+#if __linux__ && !SDL_VERSION_ATLEAST(2,0,0)
 static struct ff_effect ffeffect[4];
 static struct ff_effect ffstrong[4];
 static struct ff_effect ffweak[4];
@@ -255,7 +254,7 @@ EXPORT m64p_error CALL PluginGetVersion(m64p_plugin_type *PluginType, int *Plugi
 
 /* Helper function to handle the SDL keys */
 static void
-doSdlKeys(unsigned char* keystate)
+doSdlKeys(const unsigned char* keystate)
 {
     int c, b, axis_val, axis_max_val;
     static int grabmouse = 1, grabtoggled = 0;
@@ -306,7 +305,7 @@ doSdlKeys(unsigned char* keystate)
                     grabmouse = !grabmouse;
                     // grab/ungrab mouse
 #if SDL_VERSION_ATLEAST(2,0,0)
-#warning SDL mouse grabbing not yet supported with SDL 2.0
+                    SDL_SetRelativeMouseMode(grabmouse ? SDL_TRUE : SDL_FALSE);
 #else
                     SDL_WM_GrabInput( grabmouse ? SDL_GRAB_ON : SDL_GRAB_OFF );
 #endif
@@ -402,7 +401,15 @@ EXPORT void CALL ControllerCommand(int Control, unsigned char *Command)
                 unsigned int dwAddress = (Command[3] << 8) + (Command[4] & 0xE0);
               if (dwAddress == PAK_IO_RUMBLE && *Data)
                     DebugMessage(M64MSG_VERBOSE, "Triggering rumble pack.");
-#ifdef __linux__
+#if SDL_VERSION_ATLEAST(2,0,0)
+                if(dwAddress == PAK_IO_RUMBLE && controller[Control].event_joystick) {
+                    if (*Data) {
+                        SDL_HapticRumblePlay(controller[Control].event_joystick, 1, SDL_HAPTIC_INFINITY);
+                    } else {
+                        SDL_HapticRumbleStop(controller[Control].event_joystick);
+                    }
+                }
+#elif __linux__
                 struct input_event play;
                 if( dwAddress == PAK_IO_RUMBLE && controller[Control].event_joystick != 0)
                 {
@@ -567,9 +574,10 @@ EXPORT void CALL GetKeys( int Control, BUTTONS *Keys )
     if (controller[Control].mouse)
     {
 #if SDL_VERSION_ATLEAST(2,0,0)
-#warning SDL mouse grabbing not yet supported with SDL 2.0
+        if (SDL_GetRelativeMouseMode())
 #else
         if (SDL_WM_GrabInput(SDL_GRAB_QUERY) == SDL_GRAB_ON)
+#endif
         {
             SDL_PumpEvents();
 #if SDL_VERSION_ATLEAST(1,3,0)
@@ -578,6 +586,11 @@ EXPORT void CALL GetKeys( int Control, BUTTONS *Keys )
             while (SDL_PeepEvents(&event, 1, SDL_GETEVENT, SDL_EVENTMASK(SDL_MOUSEMOTION)) == 1)
 #endif
             {
+#if SDL_VERSION_ATLEAST(2,0,0)
+                int w, h;
+                SDL_Window *focus;
+#endif
+
                 if (event.motion.xrel)
                 {
                     mousex_residual += (int) (event.motion.xrel * controller[Control].mouse_sens[0]);
@@ -586,29 +599,42 @@ EXPORT void CALL GetKeys( int Control, BUTTONS *Keys )
                 {
                     mousey_residual += (int) (event.motion.yrel * controller[Control].mouse_sens[1]);
                 }
+
+#if SDL_VERSION_ATLEAST(2,0,0)
+                focus = SDL_GetKeyboardFocus();
+                if (focus) {
+                    SDL_GetWindowSize(focus, &w, &h);
+                    SDL_WarpMouseInWindow(focus, w / 2, h / 2);
+                } else {
+                    mousex_residual = 0;
+                    mousey_residual = 0;
+                }
+#endif
+            }
+            axis_val = mousex_residual;
+            if (axis_val < -80)
+                axis_val = -80;
+            else if (axis_val > 80)
+                axis_val = 80;
+            controller[Control].buttons.X_AXIS = axis_val;
+            axis_val = mousey_residual;
+            if (axis_val < -80)
+                axis_val = -80;
+            else if (axis_val > 80)
+                axis_val = 80;
+            controller[Control].buttons.Y_AXIS = -axis_val;
+            /* the mouse x/y values decay exponentially (returns to center), unless the left "Windows" key is held down */
+            if (!myKeyState[SDL_SCANCODE_LGUI])
+            {
+                mousex_residual = (mousex_residual * 224) / 256;
+                mousey_residual = (mousey_residual * 224) / 256;
             }
         }
         else
-#endif
         {
             mousex_residual = 0;
             mousey_residual = 0;
         }
-        axis_val = mousex_residual;
-        if (axis_val < -80)
-            axis_val = -80;
-        else if (axis_val > 80)
-            axis_val = 80;
-        controller[Control].buttons.X_AXIS = axis_val;
-        axis_val = mousey_residual;
-        if (axis_val < -80)
-            axis_val = -80;
-        else if (axis_val > 80)
-            axis_val = 80;
-        controller[Control].buttons.Y_AXIS = -axis_val;
-        /* the mouse x/y values decay exponentially */
-        mousex_residual = (mousex_residual * 224) / 256;
-        mousey_residual = (mousey_residual * 224) / 256;
     }
 
 #ifdef _DEBUG
@@ -617,7 +643,30 @@ EXPORT void CALL GetKeys( int Control, BUTTONS *Keys )
     *Keys = controller[Control].buttons;
 
     /* handle mempack / rumblepak switching (only if rumble is active on joystick) */
-#ifdef __linux__
+#if SDL_VERSION_ATLEAST(2,0,0)
+    if (controller[Control].event_joystick) {
+        static unsigned int SwitchPackTime[4] = {0, 0, 0, 0}, SwitchPackType[4] = {0, 0, 0, 0};
+        if (controller[Control].buttons.Value & button_bits[14]) {
+            SwitchPackTime[Control] = SDL_GetTicks();         // time at which the 'switch pack' command was given
+            SwitchPackType[Control] = PLUGIN_MEMPAK;          // type of new pack to insert
+            controller[Control].control->Plugin = PLUGIN_NONE;// remove old pack
+            SDL_HapticRumblePlay(controller[Control].event_joystick, 0.5, 500);
+        }
+        if (controller[Control].buttons.Value & button_bits[15]) {
+            SwitchPackTime[Control] = SDL_GetTicks();         // time at which the 'switch pack' command was given
+            SwitchPackType[Control] = PLUGIN_RAW;             // type of new pack to insert
+            controller[Control].control->Plugin = PLUGIN_NONE;// remove old pack
+            SDL_HapticRumblePlay(controller[Control].event_joystick, 1, 500);
+        }
+        // handle inserting new pack if the time has arrived
+        if (SwitchPackTime[Control] != 0 && (SDL_GetTicks() - SwitchPackTime[Control]) >= 1000)
+        {
+            SDL_HapticRumbleStop(controller[Control].event_joystick);
+            controller[Control].control->Plugin = SwitchPackType[Control];
+            SwitchPackTime[Control] = 0;
+        }
+    }
+#elif __linux__
     if (controller[Control].event_joystick != 0)
     {
         struct input_event play;
@@ -657,9 +706,66 @@ EXPORT void CALL GetKeys( int Control, BUTTONS *Keys )
     controller[Control].buttons.Value = 0;
 }
 
+static void InitiateJoysticks(int cntrl)
+{
+    // init SDL joystick subsystem
+    if (!SDL_WasInit(SDL_INIT_JOYSTICK))
+        if (SDL_InitSubSystem(SDL_INIT_JOYSTICK) == -1) {
+            DebugMessage(M64MSG_ERROR, "Couldn't init SDL joystick subsystem: %s", SDL_GetError() );
+            return;
+        }
+
+    if (controller[cntrl].device >= 0) {
+        controller[cntrl].joystick = SDL_JoystickOpen(controller[cntrl].device);
+        if (!controller[cntrl].joystick)
+            DebugMessage(M64MSG_WARNING, "Couldn't open joystick for controller #%d: %s", cntrl + 1, SDL_GetError());
+    } else {
+        controller[cntrl].joystick = NULL;
+    }
+}
+
+static void DeinitJoystick(int cntrl)
+{
+#if SDL_VERSION_ATLEAST(2,0,0)
+    if (controller[cntrl].joystick) {
+        SDL_JoystickClose(controller[cntrl].joystick);
+        controller[cntrl].joystick = NULL;
+    }
+#endif
+}
+
 static void InitiateRumble(int cntrl)
 {
-#ifdef __linux__
+#if SDL_VERSION_ATLEAST(2,0,0)
+    if (!SDL_WasInit(SDL_INIT_HAPTIC)) {
+        if (SDL_InitSubSystem(SDL_INIT_HAPTIC) == -1) {
+            DebugMessage(M64MSG_ERROR, "Couldn't init SDL haptic subsystem: %s", SDL_GetError() );
+            return;
+        }
+    }
+
+    controller[cntrl].event_joystick = SDL_HapticOpenFromJoystick(controller[cntrl].joystick);
+    if (!controller[cntrl].event_joystick) {
+        DebugMessage(M64MSG_WARNING, "Couldn't open rumble support for joystick #%i", cntrl + 1);
+        return;
+    }
+
+    if (SDL_HapticRumbleSupported(controller[cntrl].event_joystick) == SDL_FALSE) {
+        SDL_HapticClose(controller[cntrl].event_joystick);
+        controller[cntrl].event_joystick = NULL;
+        DebugMessage(M64MSG_WARNING, "Joystick #%i doesn't support rumble effect", cntrl + 1);
+        return;
+    }
+
+    if (SDL_HapticRumbleInit(controller[cntrl].event_joystick) != 0) {
+        SDL_HapticClose(controller[cntrl].event_joystick);
+        controller[cntrl].event_joystick = NULL;
+        DebugMessage(M64MSG_WARNING, "Rumble initialization failed for Joystick #%i", cntrl + 1);
+        return;
+    }
+
+    DebugMessage(M64MSG_INFO, "Rumble activated on N64 joystick #%i", cntrl + 1);
+#elif __linux__
     DIR* dp;
     struct dirent* ep;
     unsigned long features[4];
@@ -763,6 +869,16 @@ static void InitiateRumble(int cntrl)
 #endif /* __linux__ */
 }
 
+static void DeinitRumble(int cntrl)
+{
+#if SDL_VERSION_ATLEAST(2,0,0)
+    if (controller[cntrl].event_joystick) {
+        SDL_HapticClose(controller[cntrl].event_joystick);
+        controller[cntrl].event_joystick = NULL;
+    }
+#endif
+}
+
 /******************************************************************
   Function: InitiateControllers
   Purpose:  This function initialises how each of the controllers
@@ -793,10 +909,13 @@ EXPORT void CALL InitiateControllers(CONTROL_INFO ControlInfo)
     for( i = 0; i < 4; i++ )
     {
         // test for rumble support for this joystick
+        InitiateJoysticks(i);
         InitiateRumble(i);
         // if rumble not supported, switch to mempack
         if (controller[i].control->Plugin == PLUGIN_RAW && controller[i].event_joystick == 0)
             controller[i].control->Plugin = PLUGIN_MEMPAK;
+        DeinitRumble(i);
+        DeinitJoystick(i);
     }
 
     DebugMessage(M64MSG_INFO, "%s version %i.%i.%i initialized.", PLUGIN_NAME, VERSION_PRINTF_SPLIT(PLUGIN_VERSION));
@@ -833,19 +952,17 @@ EXPORT void CALL RomClosed(void)
     int i;
 
     // close joysticks
-    for( i = 0; i < 4; i++ )
-        if( controller[i].joystick )
-        {
-            SDL_JoystickClose( controller[i].joystick );
-            controller[i].joystick = NULL;
-        }
+    for( i = 0; i < 4; i++ ) {
+        DeinitRumble(i);
+        DeinitJoystick(i);
+    }
 
     // quit SDL joystick subsystem
     SDL_QuitSubSystem( SDL_INIT_JOYSTICK );
 
     // release/ungrab mouse
 #if SDL_VERSION_ATLEAST(2,0,0)
-#warning SDL mouse grabbing not yet supported with SDL 2.0
+    SDL_SetRelativeMouseMode(SDL_FALSE);
 #else
     SDL_WM_GrabInput( SDL_GRAB_OFF );
 #endif
@@ -874,23 +991,20 @@ EXPORT int CALL RomOpen(void)
         }
 
     // open joysticks
-    for( i = 0; i < 4; i++ )
-        if( controller[i].device >= 0 )
-        {
-            controller[i].joystick = SDL_JoystickOpen( controller[i].device );
-            if( controller[i].joystick == NULL )
-                DebugMessage(M64MSG_WARNING, "Couldn't open joystick for controller #%d: %s", i + 1, SDL_GetError() );
-        }
-        else
-            controller[i].joystick = NULL;
+    for (i = 0; i < 4; i++) {
+        InitiateJoysticks(i);
+        InitiateRumble(i);
+    }
 
     // grab mouse
     if (controller[0].mouse || controller[1].mouse || controller[2].mouse || controller[3].mouse)
     {
-#if SDL_VERSION_ATLEAST(2,0,0)
-#warning SDL mouse grabbing not yet supported with SDL 2.0
-#else
         SDL_ShowCursor( 0 );
+#if SDL_VERSION_ATLEAST(2,0,0)
+        if (SDL_SetRelativeMouseMode(SDL_TRUE) < 0) {
+            DebugMessage(M64MSG_WARNING, "Couldn't grab input! Mouse support won't work!");
+        }
+#else
         if (SDL_WM_GrabInput( SDL_GRAB_ON ) != SDL_GRAB_ON)
         {
             DebugMessage(M64MSG_WARNING, "Couldn't grab input! Mouse support won't work!");
